# Unit test for translation of Simulink diagrams

import unittest

from ss2hcsp.sl.sl_diagram import SL_Diagram
from ss2hcsp.sl.get_hcsp import get_hcsp, new_get_hcsp
from ss2hcsp.hcsp.simulator import SimInfo, exec_parallel
from ss2hcsp.hcsp import hcsp
from ss2hcsp.hcsp import optimize


def printTofile(path, content, module=False):
    with open(path, "w") as f:
        if module:
            f.write("%type: module\n")
        f.write(content)

def print_module(path, m):
    with open(path, "w") as f:
        f.write("%type: module\n\n")
        f.write(m.export() + '\n')
        f.write("system\n  %s=%s()\nendsystem" % (m.name, m.name))

def run_test(self, location, num_steps, expected_series, *,
             print_diagrams=False, print_hcsp_raw=False, print_hcsp=False,
             print_time_series=False, output_to_file=None, debug_name=True):
    # First, parse and process diagram
    diagram = SL_Diagram(location=location)
    diagram.parse_xml()
    diagram.delete_subsystems()
    diagram.add_line_name()
    diagram.comp_inher_st()
    diagram.inherit_to_continuous()
    diagram.separate_diagram()

    # Optional: print diagram
    if print_diagrams:
        print("Discrete blocks:")
        for block in diagram.discrete_blocks:
            print(block)
        print("Continuous blocks:")
        for block in diagram.continuous_blocks:
            print(block)
        print("Outputs:")
        print(diagram.outputs)

    # Convert to HCSP
    result_hp = new_get_hcsp(
        diagram.discrete_blocks, diagram.continuous_blocks,
        diagram.chart_parameters, diagram.outputs)

    if debug_name:
        before_size = 0
        before_size += result_hp.code.size()
        for proc in result_hp.procedures:
            before_size += proc.hp.size()

    # Optional: print HCSP before optimization
    if print_hcsp_raw:
        print(result_hp.export())

    # Optimize module
    hp = result_hp.code
    procs = dict((proc.name, proc.hp) for proc in result_hp.procedures)
    procs, hp = optimize.full_optimize_module(procs, hp)
    result_hp.procedures = [hcsp.Procedure(name, hp) for name, hp in procs.items()]
    result_hp.code = hp

    if debug_name:
        after_size = 0
        after_size += result_hp.code.size()
        for proc in result_hp.procedures:
            after_size += proc.hp.size()
        print(diagram.name, before_size, after_size)

    # Optional: print HCSP
    if print_hcsp:
        print(result_hp.export())

    # Optional: export HCSP to file
    if output_to_file:
        assert isinstance(output_to_file, str)
        print_module(output_to_file, result_hp)

    # Perform simulation
    proc_dict = dict()
    for proc in result_hp.procedures:
        proc_dict[proc.name] = proc
    info = SimInfo(result_hp.name, result_hp.code, procedures=proc_dict, outputs=result_hp.outputs)
    res = exec_parallel([info], num_steps=num_steps, verbose=False)

    # Optional: print time series
    if print_time_series:
        series = res['time_series']['P']
        for i, pt in enumerate(series):
            if i != len(series)-1 and pt['time'] == series[i+1]['time']:
                continue
            print("%.2f: %s" % (pt['time'], pt['state']))

    if expected_series:
        series = dict()
        for pt in res['time_series']['P']:
            for time in expected_series:
                if abs(pt['time'] - time) < 1e-10:
                    series[time] = pt['state']
        for time in expected_series:
            self.assertTrue(time in series, "Time %s not found in result" % time)
            self.assertTrue(len(expected_series[time]) == len(series[time]), "Unequal length at time %s" % time)
            for var in expected_series[time]:
                self.assertTrue(var in series[time])
                self.assertAlmostEqual(
                    series[time][var], expected_series[time][var],
                    msg="Disagreement at time %s, variable %s" % (time, var), places=3)

class SimTest(unittest.TestCase):
    def testLunarLander(self):
        run_test(self, "./Examples/Simulink/LunarLander.xml", 1600, {
            0: {'m': 1250, 'v': -2, 'Fc': 2437.055},
            2.020: {'m': 1248.185, 'v': -1.644, 'Fc': 2122.533},
            3.968: {'m': 1246.569, 'v': -1.564, 'Fc': 2044.469},
            6.016: {'m': 1244.903, 'v': -1.543, 'Fc': 2025.467},
            8.036: {'m': 1243.269, 'v': -1.536, 'Fc': 2019.523},
            9.984: {'m': 1241.697, 'v': -1.532, 'Fc': 2016.002}
        })

    def testMarsLander(self):
        run_test(self, "./Examples/Simulink/MarsLander.xml", 1600, {
            0.000: {'m': 759.5, 'v': -1.5, 'Fc': 2834.449},
            2.020: {'m': 757.118, 'v': -1.500, 'Fc': 2825.807},
            3.968: {'m': 754.828, 'v': -1.499, 'Fc': 2816.761},
            6.016: {'m': 752.428, 'v': -1.499, 'Fc': 2807.788},
            8.036: {'m': 750.068, 'v': -1.499, 'Fc': 2799.413},
            9.984: {'m': 747.800, 'v': -1.499, 'Fc': 2790.511}
        })

    def testVanderPol(self):
        # This test case contains blocks with different sample times
        run_test(self, "./Examples/Simulink/Van_der_Pol.xml", 50, {
            0: {'z': 1},
            1: {'z': 1.382},
            2: {'z': 0.493},
            3: {'z': -0.849},
            4: {'z': -1.410},
            5: {'z': -0.675},
            6: {'z': 0.681},
            7: {'z': 1.411},
            8: {'z': 0.844},
            9: {'z': -0.532}
        })

    def testThreeTank(self):
        run_test(self, "./Examples/Simulink/ThreeTank.xml", 600, {
            0: {'x': 0, 'y': 0, 'z': 0},
            10: {'x': 0.672, 'y': 0.028, 'z': 0.004},
            20: {'x': 1.316, 'y': 0.076, 'z': 0.014},
            30: {'x': 1.941, 'y': 0.135, 'z': 0.027},
            40: {'x': 2.552, 'y': 0.202, 'z': 0.043},
            50: {'x': 3.150, 'y': 0.277, 'z': 0.062},
            60: {'x': 3.736, 'y': 0.356, 'z': 0.083},
            70: {'x': 4.313, 'y': 0.441, 'z': 0.106},
            80: {'x': 4.880, 'y': 0.530, 'z': 0.132},
            90: {'x': 5.439, 'y': 0.623, 'z': 0.159},
            100: {'x': 5.989, 'y': 0.720, 'z': 0.188}
        })

    # def testIsolette(self):
    #     run_test(self, "./Examples/Simulink/Isolette.xml", 1200, {
    #         0.0: {'q': 75, 'c': 75},
    #         40.0: {'q': 115.0, 'c': 90.133},
    #         80.0: {'q': 103.054, 'c': 107.166},
    #         120.0: {'q': 98.645, 'c': 96.473},
    #         160.0: {'q': 89.899, 'c': 100.370},
    #         200.0: {'q': 92.852, 'c': 99.874}
    #     })

    # def testSignalBuilder(self):
    #     directory = "./Examples/signalBuilder/"
    #     xml_file = "testSignalBuilder.xml"
    #     diagram = SL_Diagram(location=directory + xml_file)
    #     model_name = diagram.parse_xml()
    #     diagram.add_line_name()
    #     diagram.comp_inher_st()
    #     diagram.inherit_to_continuous()
    #     real_hp = get_hcsp(*diagram.seperate_diagram(), model_name)
    #     printTofile(path=directory+xml_file[:-3]+"txt", content=real_hp)

    # def testVelocityControl(self):
    #     directory = "./Examples/signalBuilder/"
    #     xml_file = "velocity_control.xml"
    #     diagram = SL_Diagram(location=directory + xml_file)
    #     model_name = diagram.parse_xml()
    #     diagram.delete_subsystems()
    #     diagram.comp_inher_st()
    #     # diagram.add_buffers()
    #     diagram.add_line_name()
    #     # print(diagram)
    #     real_hp = get_hcsp(*diagram.seperate_diagram(), model_name)
    #     # print(real_hp)
    #     printTofile(path=directory+xml_file[:-3]+"txt", content=real_hp)

    def testSource1(self):
        # Value should be y = sin(t)
        run_test(self, "./Examples/Simulink/Source1.xml", 30, {
            0: {'y': 0},
            1: {'y': 0.841},
            2: {'y': 0.909},
            3: {'y': 0.141},
            4: {'y': -0.757},
            5: {'y': -0.959}
        })

    def testDelay1(self):
        run_test(self, "./Examples/Simulink/Delay1.xml", 60, {
            0: {'x': 0},
            1: {'x': 1},
            2: {'x': 2},
            3: {'x': 3},
            4: {'x': 4},
            5: {'x': 5}
        })

    def testDelay2(self):
        run_test(self, "./Examples/Simulink/Delay2.xml", 80, {
            0: {'x': 0},
            1: {'x': 0},
            2: {'x': 1},
            3: {'x': 1},
            4: {'x': 2},
            5: {'x': 2}
        })

    def testDelay3(self):
        run_test(self, "./Examples/Simulink/Delay3.xml", 60, {
            0: {'y': 1, 'x': 0},
            1: {'y': 1, 'x': 1},
            2: {'y': 2, 'x': 1},
            3: {'y': 3, 'x': 2},
            4: {'y': 5, 'x': 3},
            5: {'y': 8, 'x': 5}
        })

    # def testEnabled1(self):
    #     run_test(self, "./Examples/Simulink/Enabled1.xml", 70, {
    #
    #     })

    def testTriggered1(self):
        # Continuous triggered subsystem
        run_test(self, "./Examples/Simulink/Triggered1.xml", 70, {
            0: {'a': 0, 'y': -1},
            1: {'a': 1, 'y': 0},
            2: {'a': 1, 'y': 2},
            3: {'a': 1, 'y': 4},
            4: {'a': 1, 'y': 6},
            5: {'a': 1, 'y': 8}
        })

    def testTriggered2(self):
        # Discrete triggered subsystem
        # NOTE: the result here agrees with simulation in Matlab only if the
        # simulation time is set to 3 seconds. For longer simulation time,
        # the result from Matlab is different (and inconsistent with itself).
        run_test(self, "./Examples/Simulink/Triggered2.xml", 70, {
            0: {'z': -1, 'a': 0, 'y': -1},
            1: {'z': 0, 'a': 1, 'y': 0},
            1.5: {'z': 0, 'a': 1, 'y': 1},
            2: {'z': 2, 'a': 3, 'y': 2},
            2.5: {'z': 2, 'a': 3, 'y': 4},
            3: {'z': 6, 'a': 3, 'y': 6}
        })

    # def testTemporalLogic(self):
    #     run_test(self, "./Examples/Simulink/Temporal_Logic.xml", 80, {
    #
    #     })

    def testInputEvent(self):
        run_test(self, "./Examples/Simulink/Input_Event_2018a.xml", 80, {

        })

    def testSLIsolette(self):
        run_test(self, "./hhlpy/examples/simulink/sl_isolette.xml", 80, {

        }, output_to_file="./hhlpy/examples/simulink/sl_isolette.txt")

    def testSLDelay(self):
        run_test(self, "./hhlpy/examples/simulink/sl_delay.xml", 80, {

        }, output_to_file="./hhlpy/examples/simulink/sl_delay.txt")


if __name__ == "__main__":
    unittest.main()



"""Wrapper for SymPy"""

from decimal import Decimal
from fractions import Fraction

import sympy
from sympy.logic.boolalg import BooleanFunction, BooleanAtom, BooleanTrue, BooleanFalse

from ss2hcsp.hcsp import expr

def toSpexpr(e, functions):
    """Conversion from hcsp expression to sympy expression"""
    def rec(e):
        if isinstance(e, expr.AVar):
            return sympy.symbols(e.name)
        elif isinstance(e, expr.AConst):
            if isinstance(e.value, Decimal):
                return sympy.Rational(str(e.value))
            elif isinstance(e.value, (int, float, Fraction)):
                return sympy.sympify(e.value)
            else:
                raise NotImplementedError
        elif isinstance(e, expr.BConst):
            return sympy.sympify(e.value) 
        elif isinstance(e, expr.FunExpr):
            if len(e.exprs) == 0:          
                return sympy.symbols(str(e))
            elif e.fun_name in functions:
                return rec(expr.replace_function(e, functions))
            else:
                raise NotImplementedError
        elif isinstance(e, expr.OpExpr):
            if e.op == '+':
                return sympy.Add(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '-':
                if len(e.exprs) == 1:
                    return sympy.Mul(sympy.S.NegativeOne, rec(e.exprs[0]))
                else:
                    return sympy.Add(rec(e.exprs[0]), 
                                    sympy.Mul(sympy.S.NegativeOne, rec(e.exprs[1])))
            elif e.op == '*':
                return sympy.Mul(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '/':
                return sympy.Mul(rec(e.exprs[0]),
                                sympy.Pow(rec(e.exprs[1]), sympy.S.NegativeOne))
            elif e.op == '^':
                return sympy.Pow(rec(e.exprs[0]), rec(e.exprs[1]))
        elif isinstance(e, expr.RelExpr):
            if e.op == '<':
                return sympy.Rel(rec(e.expr1), rec(e.expr2), '<')
            elif e.op == '<=':
                return sympy.Rel(rec(e.expr1), rec(e.expr2), '<=')
            elif e.op == '>':
                return sympy.Rel(rec(e.expr1), rec(e.expr2), '>')
            elif e.op == '>=':
                return sympy.Rel(rec(e.expr1), rec(e.expr2), '>=')
            elif e.op == '==':
                return sympy.Rel(rec(e.expr1), rec(e.expr2), '==')
            elif e.op == '!=':
                return sympy.Rel(rec(e.expr1), rec(e.expr2), '!=')
        elif isinstance(e, expr.LogicExpr):
            if e.op == '&&':
                return sympy.And(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '|':
                return sympy.Or(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '!':
                return sympy.Not(rec(e.exprs[0]))
            elif e.op == '->':
                return sympy.Implies(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '<->':
                return sympy.Equivalent(rec(e.exprs[0]), rec(e.exprs[1]))
        else:
            raise NotImplementedError

    return rec(e)

def toHcsp(e):
    """Conversion from sympy expression to hcsp expression"""
    if isinstance(e, sympy.Symbol):
        return expr.AVar(e.name)
    elif isinstance(e, sympy.Number):
        if isinstance(e, sympy.Float):
            return expr.AConst(float(e))
        elif isinstance(e, sympy.Rational):
            if isinstance(e, sympy.Integer):
                return expr.AConst(int(e))
            # sympy.Raitonal(2, 3)
            else:
                return expr.AConst(Fraction(e.p, e.q))
        else:
            raise NotImplementedError
    elif isinstance(e, BooleanAtom):
        if isinstance(e, BooleanTrue):
            return expr.true_expr
        else:
            return expr.false_expr
    elif isinstance(e, sympy.Rel):
        if e.rel_op == '<':
            return expr.RelExpr('<', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.rel_op == '<=':
            return expr.RelExpr('<=', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.rel_op == '>':
            return expr.RelExpr('>', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.rel_op == '>=':
            return expr.RelExpr('>=', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.rel_op == '==':
            return expr.RelExpr('==', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.rel_op == '!=':
            return expr.RelExpr('!=', toHcsp(e.args[0]), toHcsp(e.args[1]))
    
    elif isinstance(e, sympy.Add):
        return expr.list_add(*(toHcsp(arg) for arg in e.args))
    elif isinstance(e, sympy.Mul):
        return expr.list_mul(*(toHcsp(arg) for arg in e.args))
    elif isinstance(e, sympy.Pow):
        assert len(e.args) == 2
        return expr.OpExpr('^', toHcsp(e.args[0]), toHcsp(e.args[1]))

    elif isinstance(e, BooleanFunction):
        if isinstance(e, sympy.And):
            return expr.list_conj(*(toHcsp(arg) for arg in e.args))
        elif isinstance(e, sympy.Or):
            return expr.list_disj(*(toHcsp(arg) for arg in e.args))
        elif isinstance(e, sympy.Not):
            assert len(e.args) == 1
            return expr.LogicExpr('!', toHcsp(e.args[0]))
        elif isinstance(e, sympy.Implies):
            assert len(e.args) == 2
            return expr.LogicExpr('->', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif isinstance(e, sympy.Equivalent):
            assert len(e.args) == 2
            return expr.LogicExpr('<->', toHcsp(e.args[0]), toHcsp(e.args[1]))
        else:
            raise NotImplementedError

def sp_simplify(e, functions):
    """Simplify the given expression by sympy"""
    sp_expr = toSpexpr(e, functions)
    sp_expr = sympy.simplify(sp_expr)
    hcsp_expr = toHcsp(sp_expr)

    return hcsp_expr

def sp_polynomial_div(p, q, functions):
    """Compute the quotient and remainder of polynomial p and q"""
    p = toSpexpr(p, functions)
    q = toSpexpr(q, functions)
    quot, remain = sympy.div(p, q)

    quot = toHcsp(quot)
    remain = toHcsp(remain)
    quot_remains = dict()
    quot_remains[quot] = remain

    return quot_remains

def sp_is_polynomial(e, functions, constants=set()):
    """Return True if the given expression is a polynomial and False otherwise"""
    vars = e.get_vars().difference(constants)
    vars = {sympy.Symbol(var) for var in vars}

    e = toSpexpr(e, functions)

    result = e.is_polynomial(*vars)

    return result


"""Simulink blocks."""

from decimal import Decimal
from functools import reduce
from math import gcd, pow

from ss2hcsp.sl.sl_line import SL_Line
from ss2hcsp.hcsp.expr import true_expr


def get_gcd(sample_times):
    """Return the gcd of a list of sample times.

    If some of the sample times are not integers, first multiply by an
    appropriate power of 10 before taking gcd.

    """
    if len(sample_times) == 0:
        return 0  # continuous

    assert isinstance(sample_times, (list, tuple)) and len(sample_times) >= 1
    assert all(isinstance(st, (int, Decimal)) for st in sample_times)

    if len(sample_times) == 1:
        return sample_times[0]

    if 0 in sample_times:
        return 0
    elif -2 in sample_times:
        return -2
    elif -1 in sample_times:
        return -1

    scaling_positions = []
    for st in sample_times:
        if isinstance(st, int) or int(st) == st:
            scaling_positions.append(0)
        else:  # isinstance(st, Decimal)
            scaling_positions.append(len(str(st)) - str(st).index(".") - 1)
    scale = 10 ** max(scaling_positions)
    scaling_sample_times = [int(st * scale) for st in sample_times]
    result_gcd = reduce(gcd, scaling_sample_times)
    if result_gcd % scale == 0:
        return result_gcd // int(scale)
    else:
        return Decimal(result_gcd) / scale

class SL_Block:
    """Represents a block in a Simulink diagram."""
    def __init__(self, type, name, num_src, num_dest, st):
        # Name of the block
        self.name = name

        # Type of the block, corresponds to class name
        self.type = type

        # Whether a block is continuous or discrete
        self.is_continuous = (st == 0)

        # Number of ports for lines originating from the block
        self.num_src = num_src

        # Number of ports for lines ending at the block
        self.num_dest = num_dest

        # Lines originating from the block, maintained as a list
        # of lists of SL_Line objects.
        self.src_lines = [[] for _ in range(num_src)]

        # Lines ending at the block, maintained as a list of
        # SL_Line objects
        self.dest_lines = [None] * num_dest

        # Sample time
        assert isinstance(st, (int, Decimal))
        self.st = st

        # Enabled condition
        self.enable = true_expr

    def add_src(self, port_id, sl_line):
        """Add a source line."""
        # assert port_id < self.num_src
        assert isinstance(sl_line, SL_Line)
        # First check if there is a branch that is currently None
        for branch in range(len(self.src_lines[port_id])):
            if self.src_lines[port_id][branch] is None:
                sl_line.branch = branch
                self.src_lines[port_id][branch] = sl_line
                return
        # If all filled, append at the end.
        sl_line.branch = len(self.src_lines[port_id])
        self.src_lines[port_id].append(sl_line)

    def add_dest(self, port_id, sl_line):
        """Add a destination line."""
        assert port_id < self.num_dest
        if len(self.dest_lines):
            self.dest_lines[port_id] = sl_line

    def get_src_blocks(self):
        return set(line.src for line in self.dest_lines)

    def get_input_vars(self):
        return set(line.name for line in self.dest_lines)

    def get_output_vars(self):
        assert all(len(set(line.name for line in line_list)) == 1 for line_list in self.src_lines)
        return set(line_list[0].name for line_list in self.src_lines)

    def get_init_hp(self):
        return None

    def get_output_hp(self):
        return None



"""Wrapper for Wolfram Engine."""

import os

from decimal import Decimal
from wolframclient.evaluation import WolframLanguageSession
from wolframclient.language import wl, wlexpr
from wolframclient.language.expression import WLFunction, WLSymbol
from wolframclient.exception import WolframKernelException
from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp import hcsp
import platform

# logging.basicConfig(level=logging.DEBUG)
found_wolfram = False
path = None
session = None

dirname = os.path.dirname(__file__)
filename = os.path.join(dirname, './wolframpath.txt')

if platform.system() == "Darwin":
    path = "/Applications/Wolfram Engine.app/Contents/Resources/Wolfram Player.app/Contents/MacOS/WolframKernel"
else:
    try:
        with open(filename, 'r') as f:
            path = f.readline().strip()
    except FileNotFoundError as e:
        print("Please add a file wolframpath.txt under hhlpy and place path to Wolfram Engine there.")

if path:
    try:
        session = WolframLanguageSession(path)
        found_wolfram = True
    except WolframKernelException:
        print("Failed to start Wolfram Kernel")

def toWLexpr(e, functions):
    """Convert a hcsp expression to WolframLanguage expression"""
    def rec(e):
        if isinstance(e, expr.AVar):
            return WLSymbol("Global`" + e.name)
        elif isinstance(e, expr.AConst):
            if isinstance(e.value, int):
                return e.value
            elif isinstance(e.value, Decimal):
                return wl.Rationalize(e.value)
            else:
                print(e, type(e))
                raise NotImplementedError
        elif isinstance(e, expr.FunExpr):
            if e.fun_name in functions:
                return rec(expr.replace_function(e, functions))
            return WLFunction(WLSymbol("Global`" + e.fun_name), *(rec(expr) for expr in e.exprs))
        elif isinstance(e, expr.BConst):
            if e.value is True:
                return True
            else:
                return False
        elif isinstance(e, expr.OpExpr):
            if e.op == '+':
                return wl.Plus(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '-':
                if len(e.exprs) == 1:
                    return wl.Minus(rec(e.exprs[0]))
                else:
                    return wl.Subtract(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '*':
                return wl.Times(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '/':
                return wl.Divide(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '^':
                return wl.Power(rec(e.exprs[0]), rec(e.exprs[1]))
            else:
                raise NotImplementedError

        elif isinstance(e, expr.RelExpr):
            if e.op == '>':
                return wl.Greater(rec(e.expr1), rec(e.expr2))
            elif e.op == '>=':
                return wl.GreaterEqual(rec(e.expr1), rec(e.expr2))
            elif e.op == '<':
                return wl.Less(rec(e.expr1), rec(e.expr2))
            elif e.op == '<=':
                return wl.LessEqual(rec(e.expr1), rec(e.expr2))
            elif e.op == '==':
                return wl.Equal(rec(e.expr1), rec(e.expr2))
            elif e.op == '!=':
                return wl.Unequal(rec(e.expr1), rec(e.expr2))
            else:
                raise AssertionError

        elif isinstance(e, expr.LogicExpr):
            if e.op == '&&':
                return wl.And(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '||':
                return wl.Or(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '!':
                return wl.Not(rec(e.exprs[0]))
            elif e.op == '->':
                return wl.Implies(rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op =='<->':
                return wl.Equivalent(rec(e.exprs[0]), rec(e.exprs[1]))
            else:
                raise AssertionError
        elif isinstance(e, expr.ForAllExpr):
            if isinstance(e.vars, tuple):
                return wl.Forall(wl.List(rec(var) for var in e.vars), rec(e.expr))
            else:
                return wl.ForAll(rec(e.vars), rec(e.expr))
        elif isinstance(e, expr.ExistsExpr):
            if isinstance(e.vars, tuple):
                return wl.Exists(wl.List(rec(var) for var in e.vars), rec(e.expr))
            else:
                return wl.Exists(rec(e.vars), rec(e.expr))
        else:
            raise NotImplementedError

    return rec(e)

def toHcsp(e):
    """
    Translate a wlexpression to hcsp expression or program.
    Example1:
    wl: Rule[Global`y[Global`x], Plus[Times[Rational[1, 2], Power[Global`x, 2]], C[1]]]
    hcsp: y(x) := 1/2 * x^2 + C(1)

    Example2:
    wl: Rule[Global`x[Global`t], Times[Rational[1, 2], Plus[Times[Global`a, Power[Global`t, 2]], Times[2, Global`t, Global`v0], Times[2, Global`x0]]]])
    hcsp: x(t) := 1/2 * (a * t^2 + 2 * v0 * t + 2 * x0)
    
    """
    if isinstance(e, WLFunction):
        if e.head == WLSymbol("Rule"):
            assert len(e.args) == 2
            return hcsp.Assign(toHcsp(e.args[0]), toHcsp(e.args[1]))
        
        # Translate to OpExpr in hcsp.
        elif e.head == WLSymbol("Plus"):  # priority: 65
            return expr.list_add(*(toHcsp(arg) for arg in e.args))
        elif e.head == WLSymbol("Times"):  # priority: 70
            return expr.list_mul(*(toHcsp(arg) for arg in e.args))
        elif e.head == WLSymbol("Power"):  # priority: 85
            assert len(e.args) == 2
            return expr.OpExpr('^', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.head == WLSymbol("Rational"):  # priority: 85
            assert len(e.args) == 2
            return expr.OpExpr('/', toHcsp(e.args[0]), toHcsp(e.args[1]))

        # Translate to RelExpr in hcsp.
        elif e.head == WLSymbol("Equal"):
            assert len(e.args) == 2
            return expr.RelExpr('==', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.head == WLSymbol("Greater"):
            assert len(e.args) == 2
            return expr.RelExpr('>', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.head == WLSymbol("GreaterEqual"):
            assert len(e.args) == 2
            return expr.RelExpr('>=', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.head == WLSymbol("Less"):
            assert len(e.args) == 2
            return expr.RelExpr('<', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.head == WLSymbol("LessEqual"):
            assert len(e.args) == 2
            return expr.RelExpr('<=', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.head == WLSymbol("UnEqual"):
            assert len(e.args) == 2
            return expr.RelExpr('!=', toHcsp(e.args[0]), toHcsp(e.args[1]))

        # Translate into LogicExpr in hcsp.
        elif e.head == WLSymbol("And"):
            if len(e.args) == 2:
                return expr.LogicExpr('&&', toHcsp(e.args[0]), toHcsp(e.args[1]))
            elif len(e.args) > 2:
                return expr.LogicExpr("&&", toHcsp(WLFunction(WLSymbol("And"), *e.args[:-1])),
                                            toHcsp(e.args[-1]))
            else:
                raise AssertionError
        elif e.head == WLSymbol("Or"):
            if len(e.args) == 2:
                return expr.LogicExpr('||', toHcsp(e.args[0]), toHcsp(e.args[1]))
            elif len(e.args) > 2:
                return expr.LogicExpr("||", toHcsp(WLFunction(WLSymbol("Or"), *e.args[:-1])),
                                            toHcsp(e.args[-1]))
            else:
                raise AssertionError
        elif e.head == WLSymbol("Not"):
            assert len(e.args) == 1
            return expr.LogicExpr('!', toHcsp(e.args[0]))
        elif e.head == WLSymbol("Implies"):
            assert len(e.args) == 2
            return expr.LogicExpr('->', toHcsp(e.args[0]), toHcsp(e.args[1]))
        elif e.head == WLSymbol("Equivalent"):
            assert len(e.args) == 2
            return expr.LogicExpr('<->', toHcsp(e.args[0]), toHcsp(e.args[1]))

        else:
            return expr.FunExpr(str(e.head), [toHcsp(arg) for arg in e.args])

    elif isinstance(e, WLSymbol):
        str_e = str(e)
        if str_e.startswith("Global`"):
            str_e = str_e[7:]
        return expr.AVar(str_e)
    elif isinstance(e, int):
        if isinstance(e, bool):
            return expr.BConst(e)
        else:
            return expr.AConst(e)
    else:
        assert False, "Unexpected expression: %s" % str(e)


def solveODE(hp, names, time_var):
    """ Return the solution dictionary. 

    Example: {'v': v + a * t}

    'v' : str, function name
     v + a * t : Expr

    """
    assert isinstance(hp, hcsp.ODE)
    assert isinstance(names, set)
    for name in names:
        assert isinstance(name, str)

    if not isinstance(time_var, str):
        time_var = str(time_var)

    # with WolframLanguageSession(path) as session:
    # ODEs are tranlated from hcsp form to wolfram language 
    wlexpr_eqn, init2fun = Ode2Wlexpr(hp, names, time_var)

    # Solve the differential equations.
    # For example, sln is :
    # ((Rule[Global`y[Global`x], Plus[Times[Rational[1, 2], Power[Global`x, 2]], C[1]]]))
    slns = session.evaluate(wlexpr('DSolve' + wlexpr_eqn))

    # Solutions in WLlanguage are translated into a list of Assign programs in hcsp.
    # For example, [y(x) := 1/2 * x^2 + C(1)], or 
    # [v(t) := v0 + a * t, x(t) := x0 + v0 * t + a * t^2 / 2]
    solutions = []
    for sln in slns[0]:
        solutions.append(toHcsp(sln))

    # Tranlate solution into a dictionary form and 
    # change the inital value symbol to function name symbol, for the sake of ODE solution axiom.
    # e.g. from [v(t) := v0 + a * t] to {'v' : v + a * t}
    solution_dict = dict()
    for sol in solutions:

        fun_name = str(sol.var_name.fun_name)
        solution_dict[fun_name] = sol.expr.subst(init2fun)

    return solution_dict

def Ode2Wlexpr(hp, names, time_var):
    """
    Example: for ODE <x_dot = v, v_dot = a & v > 0>
    
    Return wlexpr_eqn: [{x'[t] == v[t], v'[t] == a, x[0] == x0, v[0] == v0}, {x[t], v[t]}, t] and
           init2var  : {'x0' : x, 'v0' : v}

    """

    eqs = hp.eqs # eqs: [('x', v), ('v', a)]

    # Map the function name variable to function expression.
    var2fun = dict() # {'x' : x[t], 'v' : v[t]}
    for eq in eqs:
        var2fun[eq[0]] = expr.AVar(eq[0] + '[' + time_var + ']')

    # Differential eqautions and their inital value are recorded in the dictionary.
    ode_wleqs = dict() # {"x'[t]" : 'v[t]', "v'[t]" : 'a', 'x[0]' : x0, 'v[0]' : v0}
    # Functions that will be solved.
    tosolve_funs = [] # ['x[t]', 'v[t]']
    # Map the initial value symbol to their function name variable.
    init2var = dict()  # {'x0' : 'x', 'v0' : 'v'}

    for eq in eqs:
        x_dot_t = eq[0] + "\'" + '[' + time_var +']'  # x'[t]
        deriv_expr = str(eq[1].subst(var2fun))  # 'v[t]'
        ode_wleqs[x_dot_t] = deriv_expr # {"x'[t]" : 'v[t]'}

        tosolve_funs.append(str(var2fun[eq[0]])) # ['x[t]', 'v[t]']

        # Create a new variable for initial value symbol of each functions.
        init_var = eq[0] + '0'
        while init_var in names:
            init_var = init_var + '0'
        # {'x0': x}
        init2var[init_var] = expr.AVar(eq[0])
        x_0 = eq[0] + '[0]'
        ode_wleqs[x_0] = init_var  # {"x'[t]" : 'v[t]', 'x[0]' : x0}
     
    ode_wleqs_list = []  # ["x'[t] == v[t], "x[0] == x0"]
    for left, right in ode_wleqs.items():
        ode_wleqs_list.append(left + '==' + right)

    # [{x'[t] == v[t], v'[t] == a, x[0] == x0, v[0] == v0}, {x[t], v[t]}, t]
    wlexpr_eqn = '['+ '{' + ','.join(ode_wleqs_list) + '}' + ',' + \
                      '{' + ','.join(tosolve_funs) + '}' + ',' + \
                      time_var + \
                 ']'

    return wlexpr_eqn, init2var


def wl_prove(e, functions=dict()):
    """Attempt to prove the given condition."""
    if not isinstance(e, expr.Expr):
        raise NotImplementedError

    # If wolfram not found, just return failure
    if not found_wolfram:
        return False

    vars = e.get_vars()
    wl_expr = toWLexpr(e, functions)
    wl_vars = {toWLexpr(expr.AVar(var), functions) for var in vars}

    # wl_vars cannot be empty when using FindInstance function.
    if wl_vars:
        result = session.evaluate(wl.FindInstance(wl.Not(wl_expr), wl_vars))
        # result is an empty tuple, i.e. no instance found for the not expression.
        if not result:
            return True
        else:
            return False
    # Cases when the wl_vars is empty. For example, when wl_expr is True.
    else:
        result = session.evaluate(wl.Reduce(wl_expr, wl.Reals))
        
        if str(result) == 'True':
            return True
        else:
            return False

def wl_simplify(e, functions):
    """Simplify the given hcsp expression"""
    wl_expr = toWLexpr(e, functions)
    # Use the Simplify function in wolfram.
    wl_expr = session.evaluate(wl.Simplify(wl_expr))
    hcsp_expr = toHcsp(wl_expr)

    return hcsp_expr

def wl_polynomial_div(p, q, functions):
    """Compute the quotient and remainder of polynomial p and q"""
    vars = q.get_vars()
    vars = {toWLexpr(expr.AVar(var), functions) for var in vars}
    p = toWLexpr(p, functions)
    q = toWLexpr(q, functions)

    # result is in the form, for example, (x, 1), 
    # in which x is the quotient, 1 is the remainder.
    quot_remains = dict()
    for var in vars:
        result = session.evaluate(wl.PolynomialQuotientRemainder(p, q, var))

        quot = toHcsp(result[0])
        remain = toHcsp(result[1])

        quot_remains[quot] = remain

    return quot_remains

def wl_is_polynomial(e, functions, constants=set()):
    """Verify whether the given expression is a polynomial"""
    vars = e.get_vars().difference(constants)
    vars = {toWLexpr(expr.AVar(var), functions) for var in vars}

    e = toWLexpr(e, functions)

    result = session.evaluate(wl.PolynomialQ(e, vars))

    if str(result) == 'True':
        return True
    else:
        return False


    



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp import hcsp as hp
from ss2hcsp.hcsp.expr import AVar, FunExpr

class Mux(SL_Block):
    """docstring for Mux"""
    def __init__(self, name, inputs, ports):
        super(Mux, self).__init__(name=name, num_dest=int(inputs), num_src=1, st=-1, type="mux")
        self.inputs = inputs
        self.signal_names = list()
        self.ports=ports

    def __str__(self):
        return "%s Mux[inputs = %s]" % (self.name, str(self.inputs))

    def __repr__(self):
        return str(self)

    def get_map(self):
        out_var = self.src_lines[0][0].name
        in_vars = [line.name for line in self.dest_lines]
        return [out_var, in_vars]

    def get_expr(self):
        """Compute the assignment corresponding to a mux block."""
        in_vars = [line.name for line in self.dest_lines]
        e = AVar(in_vars[0])
        for i in range(1, len(in_vars)):
            e = FunExpr("push", [e, AVar(in_vars[i])])
        return e

    def get_output_hp(self):
        out_var = self.src_lines[0][0].name
        return hp.Assign(out_var, self.get_expr())

    def get_var_subst(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: expr}



"""Unit test for Z3 wrapper."""

import unittest

from ss2hcsp.hcsp.parser import expr_parser
from hhlpy.z3wrapper import z3_prove, convert


class Z3WrapperTest(unittest.TestCase):
    def testZ3Prove(self):
        test_case = [
            "x >= 0 --> x + 1 >= 1",
            "x >= 1 --> x + 1 >= 1",
            "x >= 1 --> x + 1 >= 1 && x >= 1",
            "x >= 0 --> x + 1 >= 1 || x >= 1",
            "ForAll x. x^2 >= 0",
            "ForAll {x, y}. x^2 + y^2 >= 0"
        ]

        for e in test_case:
            e = expr_parser.parse(e)
            self.assertTrue(z3_prove(e))
    
    def testZ3ProveFail(self):
        test_case = [
            "x >= 0 --> x >= 1",
            "x >= 0 --> x + 1 >= 1 && x >= 1",
        ]

        for e in test_case:
            e = expr_parser.parse(e)
            self.assertFalse(z3_prove(e))


if __name__ == "__main__":
    unittest.main()



from os import listdir
from os.path import isfile, join, dirname
import time

from ss2hcsp.hcsp.parser import parse_hoare_triple_with_meta
from hhlpy.hhlpy import CmdVerifier
from ss2hcsp.hcsp import expr

if __name__ == "__main__":
    start = time.perf_counter()
    path = join(dirname(__file__), "examples")
    filenames = [f for f in listdir(path) if isfile(join(path, f))]
    for file in filenames:

        if not (file.startsWith("basic") or file.startsWith("nonlinear")):
            print("Skipping {}".format(file))
            continue

        print("Running {}".format(file))
        tic = time.perf_counter()
        file = join(dirname(__file__), "examples", file)
        file = open(file,mode='r')
        file_contents = file.read()
        file.close()

        # Parse pre-condition, HCSP program, and post-condition
        hoare_triple = parse_hoare_triple_with_meta(file_contents)

        # Initialize the verifier
        verifier = CmdVerifier(
            pre=expr.list_conj(*hoare_triple.pre), 
            hp=hoare_triple.hp,
            post=expr.list_conj(*hoare_triple.post))

        # Compute wp and verify
        verifier.compute_wp()
        verifier.verify()
        toc = time.perf_counter()
        print(f"Completed in {toc - tic:0.4f} seconds")
    stop = time.perf_counter()
    print(f"All benchmarks completed in {stop - start:0.4f} seconds")



# Printing sequence diagram for a trace to a text file.


def print_sequence_diagram(trace):
    # First entry of trace contains all processes
    all_processes = sorted(trace[0]['ori_pos'].keys())

    # Mapping from process name to a list of events.
    # '|' represents no event for this process.
    events = dict()
    for process in all_processes:
        events[process] = [process]

    # List of ids
    ids = ['ID']

    num_events = 1
    def add_event(id, event, processes):
        nonlocal num_events
        num_events += 1
        ids.append(id)
        for process in all_processes:
            if process in processes:
                events[process].append(event)
            else:
                events[process].append('|')

    def add_io(id, event, inproc, outproc):
        nonlocal num_events
        num_events += 1
        ids.append(id)
        for process in all_processes:
            if process == inproc:
                events[process].append(event + '?')
            elif process == outproc:
                events[process].append(event + '!')
            else:
                events[process].append('|')

    for i, step in enumerate(trace[1:]):
        if step['type'] == 'log':
            add_event(i, step['str'], step['ori_pos'])
        elif step['type'] == 'comm':
            add_io(i, step['ch_name'], step['inproc'], step['outproc'])
        elif step['type'] == 'delay':
            add_event(i, step['str'], step['ori_pos'])
        else:
            pass

    # Maximum space for id
    max_id_space = max(len(str(id)) for id in ids) + 1

    # For each process, find the maximum space needed
    max_space = dict()
    for process in all_processes:
        max_space[process] = max(len(s) for s in events[process]) + 2

    print(max_space)

    # Now create the text file
    with open('sequence_diagram.txt', 'w', encoding='utf-8') as f:
        for i in range(num_events):
            # Print id
            f.write(' ' * (max_id_space - len(str(ids[i]))))
            f.write(str(ids[i]))
            for process in all_processes:
                text = events[process][i]
                text_len = len(events[process][i])
                space_remain = max_space[process] - text_len
                space_before = space_remain // 2
                space_after = (space_remain + 1) // 2
                f.write(' ' * space_before + text + ' ' * space_after)
            f.write('\n')



from decimal import Decimal

from ss2hcsp.sl.sl_block import SL_Block, get_gcd
from ss2hcsp.hcsp import hcsp as hp
from ss2hcsp.hcsp.expr import AVar, AConst, OpExpr, conj


class DiscretePulseGenerator(SL_Block):
    """Block for pulse generator."""
    def __init__(self, name, pulseType, amplitude, period, pulseWidth, phaseDelay):
        assert isinstance(period, Decimal)
        assert isinstance(pulseWidth, Decimal)
        assert isinstance(phaseDelay, Decimal)

        self.pulseType = pulseType
        self.amplitude = amplitude
        self.period = period
        self.pulseWidth = pulseWidth
        self.phaseDelay = phaseDelay

        # Compute sample time from period, pulseWidth and phaseDelay
        self.on_width = self.pulseWidth / 100 * self.period
        self.off_width = (1 - self.pulseWidth / 100) * self.period
        st = get_gcd([self.phaseDelay + self.on_width, self.period, self.off_width])

        # Convert start, end, and period to number of sample times
        self.start_st = int(self.phaseDelay / st)
        self.end_st = int((self.phaseDelay + self.on_width) / st)
        self.period_st = int(self.period / st)

        # Now call parent class's constructor
        super(DiscretePulseGenerator, self).__init__("DiscretePulseGenerator", name, 1, 0, st)

        # Tick variable
        self.tick_var = self.name + "_tick"

    def __str__(self):
        return "%s: DiscretePulseGenerator[amplitude = %s, period = %s, pulseWidth = %s, phaseDelay = %s]" % \
            (self.name, str(self.amplitude), str(self.period), str(self.pulseWidth), str(self.phaseDelay))

    def __repr__(self):
        return str(self)

    def get_init_hp(self):
        return hp.Assign(self.tick_var, AConst(0))

    def get_output_hp(self):
        # Output variable
        out_var = self.src_lines[0][0].name

        procs = []
        cycle_st = OpExpr("%", AVar(self.tick_var), AConst(self.period_st))
        cond = conj(hp.RelExpr(">=", cycle_st, AConst(self.start_st)),
                    hp.RelExpr("<", cycle_st, AConst(self.end_st)))
        act1 = hp.Assign(out_var, AConst(self.amplitude))
        act2 = hp.Assign(out_var, AConst(0.0))
        procs.append(hp.ITE([(cond, act1)], act2))
        procs.append(hp.Assign(self.tick_var, OpExpr("+", AVar(self.tick_var), AConst(1))))
        return hp.seq(procs)



"""Transition in Stateflow diagrams"""

from ss2hcsp.matlab.function import TransitionLabel


class Transition:
    def __init__(self, ssid, label, order=0, src="", dst=""):
        assert isinstance(label, TransitionLabel)
        self.ssid = ssid
        self.label = label
        self.order = order
        self.src = src
        self.dst = dst

    def __str__(self):
        if self.label:
            return "(%s)--%s-->(%s)" % (self.src, self.label, self.dst)
        else:
            return "(%s)--->(%s)" % (self.src, self.dst)

    def __repr__(self):
        return "Transition(%s, %s, %s, %s, %s)" % \
               (self.ssid, self.label, self.order, self.src, self.dst)



"""Classes for Matlab functions"""

class Expr:
    """Base class for Matlab (non-boolean) expressions."""
    def __init__(self):
        pass

    def priority(self):
        """Priority of expression during printing."""
        raise NotImplementedError

    def get_vars(self):
        """Returns set of variables in the expression."""
        raise NotImplementedError

    def subst(self, inst):
        """inst is a dictionary mapping variable names to expressions."""
        raise NotImplementedError


class Var(Expr):
    """Matlab variables."""
    def __init__(self, name):
        super(Var, self).__init__()
        assert isinstance(name, str)
        self.name = name

    def __repr__(self):
        return "Var(%s)" % self.name

    def __str__(self):
        return str(self.name)
    
    def __eq__(self, other):
        return isinstance(other, Var) and self.name == other.name

    def __hash__(self):
        return hash(("Var", str(self.name)))

    def priority(self):
        return 100

    def get_vars(self):
        return {self.name}

    def subst(self, inst):
        if self.name in inst:
            return inst[self.name]
        else:
            return self


class ListExpr(Expr):
    """List expressions."""
    def __init__(self, *args):
        super(ListExpr, self).__init__()
        self.args = args

    def __repr__(self):
        return "ListExpr(%s)" % (','.join(repr(arg) for arg in self.args))

    def __str__(self):
        return "[%s]" % (','.join(str(arg) for arg in self.args))

    def __eq__(self, other):
        return isinstance(other, ListExpr) and self.args == other.args

    def __hash__(self):
        return hash(("ListExpr", self.args))

    def priority(self):
        return 100

    def get_vars(self):
        return set().union(*(arg.get_vars() for arg in self.args))

    def subst(self, inst):
        return ListExpr(expr.subst(inst) for expr in self.args)


class AConst(Expr):
    """Matlab constants."""
    def __init__(self, value):
        super(AConst, self).__init__()
        assert isinstance(value, (int, float, str))
        self.value = value

    def __repr__(self):
        return "AConst(%s)" % repr(self.value)

    def __str__(self):
        if isinstance(self.value, str):
            return "\"" + self.value + "\""
        else:
            return str(self.value)

    def __eq__(self, other):
        return isinstance(other, AConst) and self.value == other.value

    def __hash__(self):
        return hash(("AConst", self.value))

    def priority(self):
        return 100

    def get_vars(self):
        return set()

    def subst(self, inst):
        return self


class OpExpr(Expr):
    """Arithmetic operations."""
    def __init__(self, op_name, *exprs):
        super(OpExpr, self).__init__()
        assert op_name in ['+', '-', '*', '/', '%']
        assert (op_name == '-' and len(exprs) == 1) or len(exprs) == 2
        self.op_name = op_name
        self.exprs = tuple(exprs)

    def __repr__(self):
        return "OpExpr(%s,%s)" % (repr(self.op_name), ','.join(repr(expr) for expr in self.exprs))

    def __str__(self):
        if self.op_name == '-' and len(self.exprs) == 1:
            s = str(self.exprs[0])
            if self.exprs[0].priority() < self.priority():
                s = '(' + s + ')'
            return '-' + s

        else:
            s1, s2 = str(self.exprs[0]), str(self.exprs[1])
            if isinstance(self.exprs[0],OpExpr) or isinstance(self.exprs[1],OpExpr):
                if self.exprs[0].priority() < self.priority():
                    s1 = '(' + s1 + ')'
                if self.exprs[1].priority() <= self.priority():
                    s2 = '(' + s2 + ')'
            return '%s %s %s' % (s1, self.op_name, s2)
    
    def __eq__(self, other):
        return isinstance(other, OpExpr) and self.op_name == other.op_name and self.exprs == other.exprs

    def __hash__(self):
        return hash(("OpExpr", self.op_name, self.exprs))

    def priority(self):
        if self.op_name == '-' and len(self.exprs) == 1:
            return 80  # priority of unary minus
        elif self.op_name in ('+', '-'):
            return 65
        elif self.op_name in ('*', '/', '%'):
            return 70
        else:
            raise NotImplementedError

    def get_vars(self):
        return set().union(*(expr.get_vars() for expr in self.exprs))

    def subst(self, inst):
        return OpExpr(self.op_name, *(expr.subst(inst) for expr in self.exprs))


class FunExpr(Expr):
    def __init__(self, fun_name, *exprs):
        super(FunExpr, self).__init__()
        assert isinstance(fun_name, str)
        self.fun_name = fun_name
        assert all(isinstance(expr, (Expr,DirectName)) for expr in exprs)
        self.exprs = tuple(exprs)

    def __repr__(self):
        return "FunExpr(%s,%s)" % (repr(self.fun_name), ",".join(repr(expr) for expr in self.exprs))

    def __str__(self):
        return "%s(%s)" % (self.fun_name, ",".join(str(expr) for expr in self.exprs))

    def __eq__(self, other):
        return isinstance(other, FunExpr) and self.fun_name == other.fun_name and \
               self.exprs == other.exprs

    def __hash__(self):
        return hash(("FunExpr", self.fun_name, self.exprs))

    def priority(self):
        return 100

    def get_vars(self):
        return set().union(*(expr.get_vars() for expr in self.exprs))

    def subst(self, inst):
        return FunExpr(self.fun_name, *(expr.subst(inst) for expr in self.exprs))


class BExpr:
    """Base class for Matlab boolean expressions."""
    def __init__(self):
        pass

    def priority(self):
        raise NotImplementedError

    def get_vars(self):
        raise NotImplementedError

    def subst(self, inst):
        raise NotImplementedError


class BConst(BExpr):
    """Boolean constants (true or false)."""
    def __init__(self, value):
        super(BConst, self).__init__()
        assert isinstance(value, bool)
        self.value = value

    def __repr__(self):
        return "BConst(%s)" % str(self.value)

    def __str__(self):
        return "true" if self.value else "false"

    def __eq__(self, other):
        return isinstance(other, BConst) and self.value == other.value

    def __hash__(self):
        return hash(("BConst", self.value))

    def priority(self):
        return 100

    def get_vars(self):
        return set()

    def subst(self, inst):
        return self


class LogicExpr(BExpr):
    """Boolean operations."""
    def __init__(self, op_name, *exprs):
        super(LogicExpr, self).__init__()
        assert op_name in ["~", "&&", "||", "-->", "<-->"]
        assert (op_name == "~" and len(exprs) == 1) or (op_name != "~" and len(exprs) == 2)
       
        self.op_name = op_name
        self.exprs = tuple(exprs)
        self.expr1=self.exprs[0]
        if len(exprs) != 1:
            self.expr2=self.exprs[1]
        

    def __repr__(self):
        return "Logic(%s,%s)" % (repr(self.op_name), ','.join(repr(expr) for expr in self.exprs))

    def __str__(self):
        if self.op_name == '~':
            s = str(self.expr1)
            if self.expr1.priority() < self.priority():
                s = '(' + s + ')'
            return '~' + s

        else:
            s1, s2 = str(self.expr1), str(self.expr2)
            if isinstance(self.expr1,LogicExpr) or isinstance(self.expr2,LogicExpr):
                if self.expr1.priority() < self.priority():
                    s1 = '(' + s1 + ')'
                if self.expr2.priority() <= self.priority():
                    s2 = '(' + s2 + ')'
            return '%s %s %s' % (s1, self.op_name, s2)

    def __eq__(self, other):
        return isinstance(other, LogicExpr) and self.op_name == other.op_name and self.exprs == other.exprs

    def __hash__(self):
        return hash(("LogicExpr", self.op_name, self.exprs))

    def priority(self):
        if self.op_name == '~':
            return 40
        elif self.op_name == '&&':
            return 35
        elif self.op_name == '||':
            return 30
        elif self.op_name == '<-->':
            return 25
        elif self.op_name == '-->':
            return 20
        else:
            raise NotImplementedError

    def get_vars(self):
        return set().union(*(expr.get_vars() for expr in self.exprs))

    def subst(self, inst):
        return LogicExpr(self.op_name, *(expr.subst(inst) for expr in self.exprs))


class RelExpr(BExpr):
    neg_table = {"<": ">=", ">": "<=", "==": "!=", "!=": "==", ">=": "<", "<=": ">"}

    def __init__(self, op, expr1, expr2):
        super(RelExpr, self).__init__()
        assert op in ["<", ">", "==", "!=", ">=", "<="]
        assert isinstance(expr1, (Expr,FunctionCall,DirectName)) and isinstance(expr2, (Expr,FunctionCall,AConst))
        self.op = op
        self.expr1 = expr1
        self.expr2 = expr2

    def __repr__(self):
        return "Rel(%s,%s,%s)" % (repr(self.op), repr(self.expr1), repr(self.expr2))

    def __str__(self):
        return str(self.expr1) + " " + self.op + " " + str(self.expr2)

    def __eq__(self, other):
        return isinstance(other, RelExpr) and self.op == other.op and \
            self.expr1 == other.expr1 and self.expr2 == other.expr2

    def __hash__(self):
        return hash(("RelExpr", self.op, self.expr1, self.expr2))

    def priority(self):
        return 50

    def get_vars(self):
        return self.expr1.get_vars().union(self.expr2.get_vars())

    def subst(self, inst):
        return RelExpr(self.op, self.expr1.subst(inst), self.expr2.subst(inst))


class Command:
    """Base class for Matlab commands."""
    def __init__(self):
        pass

    def subst(self, inst):
        """inst is a dictionary mapping variable names to expressions."""
        raise NotImplementedError


class Skip(Command):
    """Skip command."""
    def __init__(self):
        super(Skip, self).__init__()

    def __str__(self):
        return "skip"

    def __repr__(self):
        return "Skip()"

    def __eq__(self, other):
        return isinstance(other, Skip)

    def __hash__(self):
        return hash("SKIP")

    def priority(self):
        return 100

    def subst(self, inst):
        return self


class Assign(Command):
    """Assignment command.

    This command evaluates expr, which should return either a value or a list
    of values. The resulting value is assigned to one or more lnames.

    """
    def __init__(self, lname, expr):
        super(Assign, self).__init__()
        assert isinstance(expr, Expr)
        assert isinstance(lname, (Var, FunExpr, ListExpr, DirectName))
        self.lname = lname
        self.expr = expr

    def __str__(self):
        return "%s = %s" % (self.lname, self.expr)

    def __repr__(self):
        return "Assign(%s,%s)" % (repr(self.lname), repr(self.expr))

    def __eq__(self, other):
        return isinstance(other, Assign) and self.lname == other.lname and \
            self.expr == other.expr

    def __hash__(self):
        return hash(("ASSIGN", self.lname, self.expr))

    def priority(self):
        return 100

    def contain_hp(self, name):
        return False

    def subst(self, inst):
        # Disallow instantiating lname
        assert self.lname not in inst, "Cannot instantiate lname %s" % self.lname
        return Assign(self.lname, self.expr.subst(inst))


class FunctionCall(Command):
    """Function call as a single command.
    
    This includes both Matlab functions and graphical functions defined in
    Stateflow diagrams.

    """
    def __init__(self, func_name, *args):
        super(FunctionCall, self).__init__()
        self.func_name = func_name
        self.args = tuple(args)

    def __str__(self):
        return "%s(%s)" % (self.func_name, ','.join(str(arg) for arg in self.args))

    def __repr__(self):
        if self.args:
            return "FunctionCall(%s,%s)" % (repr(self.func_name), ','.join(repr(arg) for arg in self.args))
        else:
            return "FunctionCall(%s)" % repr(self.func_name)

    def __eq__(self, other):
        return isinstance(other, FunctionCall) and self.func_name == other.func_name and \
            self.args == other.args

    def __hash__(self):
        return hash(("FUNCTION_CALL", self.func_name, self.args))

    def subst(self, inst):
        return FunctionCall(self.func_name, *(arg.subst(inst) for arg in self.args))


class Sequence(Command):
    """Sequential commands."""
    def __init__(self, cmd1, cmd2):
        super(Sequence, self).__init__()
        assert isinstance(cmd1, Command) and isinstance(cmd2, Command)
        self.cmd1 = cmd1
        self.cmd2 = cmd2

    def __str__(self):
        return "%s; %s" % (str(self.cmd1), Sequence(self.cmd2.cmd1,self.cmd2.cmd2) if isinstance(self.cmd2,Sequence) else str(self.cmd2))

    def __repr__(self):
        return "Sequence(%s,%s)" % (repr(self.cmd1), Sequence(self.cmd2.cmd1,self.cmd2.cmd2) if isinstance(self.cmd2,Sequence) else repr(self.cmd2) )

    def __eq__(self, other):
        return isinstance(other, Sequence) and self.cmd1 == other.cmd1 and self.cmd2 == other.cmd2

    def __hash__(self):
        return hash(("SEQUENCE", self.cmd1, self.cmd2))

    def subst(self, inst):
        return Sequence(self.cmd1.subst(inst), self.cmd2.subst(inst))

    def priority(self):
        return 70

    def contain_hp(self, name):
        return False

def flatten(args):
    """Expand any Sequence in list of commands."""
    res = []
    for arg in args:
        if isinstance(arg, Sequence):
            res.extend(flatten([arg.cmd1]))
            res.extend(flatten([arg.cmd2]))
        else:
            res.append(arg)
    return res

def rec_seq(args):
    if len(args) == 0:
        return Skip()
    elif len(args) == 1:
        return args[0]
    else:
        return Sequence(args[0], rec_seq(args[1:]))

def seq(args):
    args = [arg for arg in flatten(args) if arg != Skip()]
    return rec_seq(args)


class IfElse(Command):
    """If-Else commands."""
    def __init__(self, cond, cmd1, cmd2):
        super(IfElse, self).__init__()
        assert isinstance(cond, BExpr) and isinstance(cmd1, Command) 
        self.cond = cond
        self.cmd1 = cmd1
        self.cmd2 = cmd2

    def __str__(self):
        if_str = str(self.cmd1)
        else_str = str(self.cmd2) if self.cmd2 is not None else ""
        res = "if %s " % self.cond
        res += ''.join(' ' + line for line in if_str.split('\n'))
        if  self.cmd2 is not None:
            res += ' else'
            res += ''.join(' ' + line for line in else_str.split('\n'))
        return res

    def __repr__(self):
        return "IfElse(%s,%s,%s)" % (repr(self.cond), repr(self.cmd1), repr(self.cmd2))

    def __eq__(self, other):
        return isinstance(other, IfElse) and self.cond == other.cond and self.cmd1 == other.cmd1 and \
            self.cmd2 == other.cmd2

    def __hash__(self):
        return hash(("IFELSE", self.cond, self.cmd1, self.cmd2))

    def subst(self, inst):
        return IfElse(self.cond.subst(inst), self.cmd1.subst(inst), self.cmd2.subst(inst)) if self.cmd2 is not None \
                else IfElse(self.cond.subst(inst), self.cmd1.subst(inst), None)


class Event:
    """Base class for events in Matlab."""
    pass

class BroadcastEvent(Event):
    """Broadcast event in Matlab.
    
    Event can serve as either conditions (require this event to be active)
    or as commands (raise this event).
    
    """
    def __init__(self, name):
        assert isinstance(name, str)
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return "BroadcastEvent(%s)" % repr(self.name)

    def __eq__(self, other):
        return isinstance(other, BroadcastEvent) and self.name == other.name

    def __hash__(self):
        return hash(("BROADCAST_EVENT", self.name))

class DirectedEvent(Event):
    """Sending event to particular state."""
    def __init__(self, state_name, event):
        assert isinstance(state_name, str) and isinstance(event, (BroadcastEvent, DirectedEvent))
        self.state_name = state_name
        self.event = event

    def __str__(self):
        return self.state_name + "." + str(self.event)

    def __repr__(self):
        return "DirectedEvent(%s,%s)" % (repr(self.state_name), repr(self.event))

    def __eq__(self, other):
        return isinstance(other, DirectedEvent) and self.state_name == other.state_name and \
            self.event == other.event

    def __hash__(self):
        return hash(("DIRECTED_EVENT", self.state_name, self.event))

class ImplicitEvent(Event):
    """Implicit events in Matlab."""
    def __init__(self, name):
        assert name in ('tick', 'wakeup')
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return "ImplicitEvent(%s)" % repr(self.name)

    def __eq__(self, other):
        return isinstance(other, ImplicitEvent) and self.name == other.name

    def __hash__(self):
        return hash(("IMPLICIT_EVENT", self.name))

class AbsoluteTimeEvent(Event):
    """Absolute time events in Matlab."""
    def __init__(self, name):
        assert name in ('sec', 'msec', 'usec')
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return "AbsoluteTimeEvent(%s)" % repr(self.name)

    def __eq__(self, other):
        return isinstance(other, AbsoluteTimeEvent) and self.name == other.name

    def __hash__(self):
        return hash(("ABSOLUTE_TIME_EVENT", self.name))

class TemporalEvent(Event):
    """Temporal logic events in Matlab."""
    def __init__(self, temp_op, expr, event):
        assert temp_op in ('after', 'before', 'at', 'every')
        assert isinstance(event, Event)
        # assert isinstance(expr, Expr)
        self.temp_op = temp_op
        self.expr = expr
        self.event = event

    def __str__(self):
        return "%s(%s,%s)" % (self.temp_op, str(self.expr), str(self.event))

    def __repr__(self):
        return "TemporalEvent(%s,%s,%s)" % (repr(self.temp_op), repr(self.expr), repr(self.event))

    def __eq__(self, other):
        return isinstance(other, TemporalEvent) and self.temp_op == other.temp_op and \
            self.expr == other.expr and self.event == other.event

    def __hash__(self):
        return hash(("TEMPORAL_EVENT", self.temp_op, self.expr, self.event))


class RaiseEvent(Command):
    """Command for raising an event."""
    def __init__(self, event):
        assert isinstance(event, (BroadcastEvent, DirectedEvent))
        self.event = event

    def __str__(self):
        return str(self.event)

    def __repr__(self):
        return "RaiseEvent(%s)" % repr(self.event)

    def __eq__(self, other):
        return isinstance(other, RaiseEvent) and self.event == other.event

    def __hash__(self):
        return hash(("RAISE_EVENT", self.event))

    def subst(self, inst):
        return self


class Message(Command):
    """Command for raising an event."""
    def __init__(self, message):
        assert isinstance(message, str)
        self.message = str(message)

    def __str__(self):
        return self.message

    def __repr__(self):
        return "RaiseEvent(%s)" % repr(self.message)

    def __eq__(self, other):
        return self.message == other.message

    def __hash__(self):
        return hash(("MESSAGE", self.message))

    def subst(self, inst):
        return self


class Function:
    """Function declarations in Matlab.
    
    name : str - name of the function.
    params : List[str] - list of parameter names.
    cmd : Command - body of the function.
    return_var : [str, List[str]] - return variables.

    """
    def __init__(self, name, params, cmd, return_var):
        assert isinstance(cmd, Command)
        self.name = name
        self.params = tuple(params)
        self.cmd = cmd
        self.type = "MATLAB_FUNCTION"
        if return_var is None or isinstance(return_var, str):
            self.return_var = return_var
        else:
            self.return_var = tuple(return_var)

    def __str__(self):
        if self.return_var is None:
            str_return_var = ""
        elif isinstance(self.return_var, str):
            str_return_var = self.return_var + "="
        else:
            str_return_var = "[" + ','.join(self.return_var) + "]="

        func_sig = str_return_var + self.name + "(" + ",".join(self.params) + ")"
        str_cmd = '\n'.join('  ' + line for line in str(self.cmd).split('\n'))
        return "function %s\n%s" % (func_sig, str_cmd)

    def __repr__(self):
        return "Function(%s,%s,%s,%s)" % (repr(self.name), repr(self.params), repr(self.cmd), repr(self.return_var))

    def __eq__(self, other):
        return isinstance(other, Function) and self.name == other.name and self.params == other.params and \
            self.cmd == other.cmd and self.return_var == other.return_var

    
    def get_Sequence(self,lists):
        if len(lists) == 1:
            return lists[0]
        else:
           return Sequence(lists[0],Sequence(self.get_Sequence(lists[1:])))
    def instantiate(self, vals=None):
        """Instantiate a procedure with given values for parameters.

        vals : [None, List[Expr]] - list of expressions as input values.
            Default to None if no input values.

        Returns Command - instantiated command.
        
        This works by replacing occurrence of parameters in the body of
        the function with the given values.
 
        """
        if vals is None:
            vals = tuple()
        assert len(self.params) == len(vals), "Function instantiation: wrong number of inputs"
        inst = dict(zip(self.params, vals))
       
        paramsList=list()
        for i in range(0,len(self.params)):
            paramsList.append(Assign(Var(self.params[i]),vals[i]))
        # return self.cmd.subst(inst),list()
        return self.cmd,paramsList


class TransitionLabel:
    """Label on transitions.

    There are four parts to specify a transition:
    event : [None, Event] - event trigger for the transition.
    cond : [None, BExpr] - condition trigger for the transition.
    cond_act : [None, Command] - action performed when checking conditions.
    tran_act : [None, Command] - action performed at the end of transition.

    """
    def __init__(self, event, cond, cond_act, tran_act):
        assert event is None or isinstance(event, Event)
        assert cond is None or isinstance(cond, (BExpr,FunExpr))
        if cond_act is None:
            cond_act = Skip()
        if tran_act is None:
            tran_act = Skip()
        assert isinstance(cond_act, Command)
        assert isinstance(tran_act, Command)

        self.event = event
        self.cond = cond
        self.cond_act = cond_act
        self.tran_act = tran_act

    def __str__(self):
        event_str = str(self.event) if self.event else ""
        cond_str = '[' + str(self.cond) + ']' if self.cond else ""
        cond_act_str = '{' + str(self.cond_act) + '}' if self.cond_act != Skip() else ""
        tran_act_str = '/{' + str(self.tran_act) + '}' if self.tran_act != Skip() else ""
        return event_str + cond_str + cond_act_str + tran_act_str

    def __repr__(self):
        return "TransitionLabel(%s,%s,%s,%s)" % (
            repr(self.event), repr(self.cond), repr(self.cond_act), repr(self.tran_act))

    def __eq__(self, other):
        return isinstance(other, TransitionLabel) and self.event == other.event and \
            self.cond == other.cond and self.cond_act == other.cond_act and self.tran_act == other.tran_act

class DirectName:
    """Directed name."""
    def __init__(self, expr):
        self.exprs = expr

    def __str__(self):
        return str(".".join(str(arg) for arg in self.exprs))

    def __repr__(self):
        return "str(%s)" %(".".join(repr(arg) for arg in self.exprs))


class StateInnerOperate:
    def __init__(self, name, op):
        self.name = name
        self.op = op

    def __str__(self):
        return self.name + ":" + self.op

    def __repr__(self):
        return "StateInnerOperate(%s,%s)" % (self.name, self.op)

class StateOperate:
    def __init__(self, name, en_op, du_op, ex_op):
        self.name = name
        self.en_op = en_op
        self.du_op = du_op
        self.ex_op = ex_op

    def __str__(self):
        state_name = self.name + "\n"
        state_en_op = self.en_op.name + " : " + self.op + "\n"
        state_du_op = self.du_op.name + " : " + self.op + "\n"
        state_ex_op = self.ex_op.name + " : " + self.op + "\n"
        return state_name + state_en_op + state_du_op + state_ex_op

    def __repr__(self):
        return "StateOperate(%s,%s,%s,%s)" % (self.name, self.en_op, self.du_op, self.ex_op)



"""Simulink blocks DataStoreMemory and DataStoreRead."""

from ss2hcsp.hcsp import hcsp 
from ss2hcsp.hcsp.expr import AVar, AConst
from ss2hcsp.sl.sl_block import SL_Block


class DataStoreMemory:
	"""Simulink block DataStoreMemory."""
	def __init__(self, name, value, dataStoreName):
		super(DataStoreMemory, self).__init__()
		self.type = "DataStoreMemory"
		self.name = name
		self.value = value
		self.dataStoreName = dataStoreName
		self.st = -1
		self.src_lines = [[]]
		self.dest_lines = [None]

	def __str__(self):
		return "name=%s,value=%s,dataStoreName=%s" % (
			self.name, str(self.value), self.dataStoreName)

	def __repr__(self):
		return "DataStoreMemory(%s,%s,%s)" % (self.name, str(self.value), self.dataStoreName)
	
	def get_hcsp(self):
		# hcsp.SelectComm(((hcsp.InputChannel("ch_"+self.dataStore_name, AVar(self.dataStore_name)),hcsp.Skip()),(hcsp.OutputChannel("ch_"+self.dataStore_name, AVar(self.dataStore_name)),hcsp.Skip())))
		# return hcsp.Sequence(hcsp.Assign(AVar(self.dataStore_name), AConst(self.value)),hcsp.Loop(hcsp.SelectComm((hcsp.OutputChannel("ch_"+self.dataStore_name, AVar(self.dataStore_name)),hcsp.Skip()),(hcsp.InputChannel("ch_"+self.dataStore_name, AVar(self.dataStore_name)),hcsp.Skip()))))
		hp_body = [
			hcsp.OutputChannel("ch_"+self.dataStoreName, AVar(self.dataStoreName)),
			hcsp.InputChannel("ch_"+self.dataStoreName, AVar(self.dataStoreName))
		]
		return hcsp.Sequence(
			hcsp.Assign(AVar(self.dataStoreName), AConst(self.value)),
			hcsp.OutputChannel("ch_"+self.dataStoreName, AVar(self.dataStoreName)),
			hcsp.InputChannel("ch_"+self.dataStoreName, AVar(self.dataStoreName)),
			hcsp.Loop(hcsp.Sequence(*hp_body))
		)


class DataStoreRead(SL_Block):
	"""Simulink block DataStoreRead."""
	def __init__(self, name, dataStoreName):
		super(DataStoreRead, self).__init__("DataStoreRead", name, 1, 1, -1)
		self.dataStoreName = dataStoreName
		self.data_length = 0

	def __str__(self):
		return "name=%s,dataStore_name=%s" % (self.name, self.dataStoreName)

	def __repr__(self):
		return "DataStoreRead(%s,%s)" % (self.name, self.dataStoreName)

	def get_hcsp(self, j):
		hp_body = []
		for i in range(j):
			var = "value" + str(i + 1)
			hp_body.append(hcsp.InputChannel("ch_"+self.dataStore_name+"_"+var, AVar(var)))
		return hcsp.Sequence(hcsp.Loop(hcsp.Sequence(*hp_body)))



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, FunExpr, true_expr, RelExpr, AConst
from ss2hcsp.hcsp import hcsp as hp


class Abs(SL_Block):
    """Compute the absolute value of the dest line."""
    def __init__(self, name, st=-1):
        super(Abs, self).__init__("abs", name, 1, 1, st)

    def get_expr(self):
        in_var = AVar(self.dest_lines[0].name)
        return FunExpr("abs", [in_var])
        
    def __str__(self):
        out_var = self.src_lines[0][0].name
        expr = self.get_expr()
        return "%s: %s = %s  (st = %s)" % (self.name, out_var, expr, self.st)

    def __repr__(self):
        return "Abs(%s, %s, in = %s, out = %s)" % \
            (self.name, self.st, self.dest_lines, self.src_lines)

    def get_output_hp(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return hp.Assign(out_var, expr)

    def get_var_subst(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: expr}

    def get_var_map(self):
        in_var = AVar(self.dest_lines[0].name)
        expr = FunExpr("abs", [in_var])
        out_var = self.src_lines[0][0].name
        return {out_var: [(true_expr, expr)]}
